import os.path
from nipype.interfaces.base import (
    CommandLineInputSpec, CommandLine, File, Directory, TraitedSpec, isdefined,
    traits, InputMultiPath)
from nipype.interfaces.mrtrix3.reconst import (
    MRTrix3Base, MRTrix3BaseInputSpec)
from arcana.utils import split_extension


# =============================================================================
# MR Convert
# =============================================================================


class MRConvertInputSpec(MRTrix3BaseInputSpec):
    in_file = traits.Either(
        File(exists=True, desc="Input file"),
        Directory(exists=True, desc="Input directory (assumed to be DICOM)"),
        mandatory=True, argstr='%s', position=-2)
    out_file = File(
        genfile=True, argstr='%s', position=-1, hash_files=False,
        desc=("Output (converted) file. If no path separators (i.e. '/' on "
              "*nix) are found in the provided output file then the CWD (when "
              "the workflow is run, i.e. the working directory) will be "
              "prepended to the output path."))
    out_ext = traits.Str(
        mandatory=False,
        desc=("The extension (and therefore the file format) to use when the "
              "output file path isn't provided explicitly"))
    coord = traits.Tuple(
        traits.Int(), traits.Int(),
        mandatory=False, argstr='-coord %d %d',
        desc=("extract data from the input image only at the coordinates "
              "specified."))
    vox = traits.Str(
        mandatory=False, argstr='-vox %s',
        desc=("change the voxel dimensions of the output image. The new sizes "
              "should be provided as a comma-separated list of values. Only "
              "those values specified will be changed. For example: 1,,3.5 "
              "will change the voxel size along the x & z axes, and leave the "
              "y-axis voxel size unchanged."))
    axes = traits.Str(
        mandatory=False, argstr='-axes %s',
        desc=("specify the axes from the input image that will be used to form"
              " the output image. This allows the permutation, ommission, or "
              "addition of axes into the output image. The axes should be "
              "supplied as a comma-separated list of axes. Any ommitted axes "
              "must have dimension 1. Axes can be inserted by supplying -1 at "
              "the corresponding position in the list."))
    scaling = traits.Str(
        mandatory=False, argstr='-scaling %s',
        desc=("specify the data scaling parameters used to rescale the "
              "intensity values. These take the form of a comma-separated "
              "2-vector of floating-point values, corresponding to offset & "
              "scale, with final intensity values being given by offset + "
              "scale * stored_value. By default, the values in the input "
              "image header are passed through to the output image header "
              "when writing to an integer image, and reset to 0,1 (no scaling)"
              " for floating-point and binary images. Note that his parameter has"
              " no effect for floating-point and binary images."))
    stride = traits.Str(
        mandatory=False, argstr='-stride %s',
        desc=("specify the strides of the output data in memory, as a "
              "comma-separated list. The actual strides produced will depend "
              "on whether the output image format can support it."))
    dataset = traits.Str(
        mandatory=False, argstr='-dataset %s',
        desc=("specify output image data type. Valid choices are: float32, "
              "float32le, float32be, float64, float64le, float64be, int64, "
              "uint64, int64le, uint64le, int64be, uint64be, int32, uint32, "
              "int32le, uint32le, int32be, uint32be, int16, uint16, int16le, "
              "uint16le, int16be, uint16be, cfloat32, cfloat32le, cfloat32be, "
              "cfloat64, cfloat64le, cfloat64be, int8, uint8, bit."))
    export_grad_mrtrix = traits.Str(
        mandatory=False, argstr='-export_grad_mrtrix %s',
        desc=("export the diffusion-weighted gradient table to file in MRtrix "
              "format"))
    export_grad_fsl = traits.Str(
        mandatory=False, argstr='-export_grad_fsl %s',
        desc=("export the diffusion-weighted gradient table to files in FSL "
              "(bvecs / bvals) format"))
    quiet = traits.Bool(
        mandatory=False, argstr="-quiet",
        desc="Don't display output during conversion")


class MRConvertOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='Extracted encoding gradients')


class MRConvert(MRTrix3Base):

    _cmd = 'mrconvert'
    input_spec = MRConvertInputSpec
    output_spec = MRConvertOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            base, orig_ext = split_extension(
                os.path.basename(self.inputs.in_file))
            ext = (self.inputs.out_ext
                   if isdefined(self.inputs.out_ext) else orig_ext)
            out_name = os.path.join(os.getcwd(),
                                    "{}_conv{}".format(base, ext))
        return out_name


class MRCatInputSpec(CommandLineInputSpec):

    first_scan = traits.File(
        exists=True, mandatory=True, desc="First input image", argstr="%s",
        position=-3)

    second_scan = traits.File(
        exists=True, mandatory=True, desc="Second input image", argstr="%s",
        position=-2)

    out_file = traits.File(
        genfile=True, desc="Output filename", position=-1, hash_files=False,
        argstr="%s")

    axis = traits.Int(
        desc="The axis along which the scans will be concatenated",
        argstr="-axis %s")

    quiet = traits.Bool(
        mandatory=False, argstr="-quiet",
        desc="Don't display output during concatenation")


class MRCatOutputSpec(TraitedSpec):

    out_file = File(exists=True, desc='Pre-processed DWI dataset')


class MRCat(CommandLine):

    _cmd = 'mrcat'
    input_spec = MRCatInputSpec
    output_spec = MRCatOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            gen_name = self._gen_outfilename()
        else:
            assert False
        return gen_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            out_name = self.inputs.out_file
        else:
            first, ext = split_extension(
                os.path.basename(self.inputs.first_scan))
            second, _ = split_extension(
                os.path.basename(self.inputs.second_scan))
            out_name = os.path.join(
                os.getcwd(), "{}_{}_concat{}".format(first, second, ext))
        return out_name


# =============================================================================
# MR Crop
# =============================================================================

class MRCropInputSpec(CommandLineInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-2,
                   desc="Diffusion weighted images with graident info")

    out_file = File(genfile=True, argstr='%s', position=-1,
                    desc="Extracted DW or b-zero images")

    axis = traits.Tuple(
        traits.Int(desc="index"),  # @UndefinedVariable
        traits.Int(desc="start"),  # @UndefinedVariable
        traits.Int(desc='end'),  # @UndefinedVariable
        mandatory=False, argstr="-axis %s %s %s", # @UndefinedVariable @IgnorePep8
        desc=("crop the input image in the provided axis"))

    mask = File(mandatory=False, exists=True, argstr="-mask %s",
                desc=("Crop the input image according to the spatial extent of"
                      " a mask image"))

    quiet = traits.Bool(
        mandatory=False, argstr="-quiet",
        desc="Don't display output during operation")


class MRCropOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='The resultant image')


class MRCrop(CommandLine):
    """
    Extracts the gradient information in MRtrix format from a DWI image
    """
    _cmd = 'mrcrop'
    input_spec = MRCropInputSpec
    output_spec = MRCropOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = self._gen_outfilename()
        else:
            assert False
        return fname

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            filename = self.inputs.out_file
        else:
            base, ext = split_extension(os.path.basename(self.inputs.in_file))
            filename = os.path.join(os.getcwd(),
                                    "{}_crop{}".format(base, ext))
        return filename


# =============================================================================
# MR Pad
# =============================================================================

class MRPadInputSpec(CommandLineInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-2,
                   desc="Diffusion weighted images with graident info")

    out_file = File(genfile=True, argstr='%s', position=-1,
                    desc="Extracted DW or b-zero images")

    axis = traits.Tuple(
        traits.Int(desc="index"),  # @UndefinedVariable
        traits.Int(desc="lower"),  # @UndefinedVariable
        traits.Int(desc='upper'),  # @UndefinedVariable
        mandatory=False, argstr="-axis %s %s %s", # @UndefinedVariable @IgnorePep8
        desc=("Pad the input image along the provided axis (defined by index)."
              "Lower and upper define the number of voxels to add to the lower"
              " and upper bounds of the axis"))

    uniform = File(mandatory=False, exists=True, argstr="-uniform %s",
                   desc=("Pad the input image by a uniform number of voxels on"
                         " all sides (in 3D)"))

    quiet = traits.Bool(
        mandatory=False, argstr="-quiet",
        desc="Don't display output during operation")


class MRPadOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='The resultant image')


class MRPad(CommandLine):
    """
    Extracts the gradient information in MRtrix format from a DWI image
    """
    _cmd = 'mrpad'
    input_spec = MRPadInputSpec
    output_spec = MRPadOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = self._gen_outfilename()
        else:
            assert False
        return fname

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            filename = self.inputs.out_file
        else:
            base, ext = split_extension(os.path.basename(self.inputs.in_file))
            filename = os.path.join(os.getcwd(),
                                    "{}_pad{}".format(base, ext))
        return filename


# =============================================================================
# MR math
# =============================================================================

class MRMathInputSpec(CommandLineInputSpec):

    in_files = InputMultiPath(
        File(exists=True), argstr='%s', mandatory=True,
        position=3, desc="Diffusion weighted images with graident info")

    out_file = File(genfile=True, argstr='%s', position=-1,
                    desc="Extracted DW or b-zero images")

    operation = traits.Str(mandatory=True, argstr='%s', position=-2,  # @UndefinedVariable @IgnorePep8
                           desc=("Operation to apply to the files"))

    axis = traits.Int(argstr="-axis %s", position=0,  # @UndefinedVariable @IgnorePep8
                      desc=("The axis over which to apply the operator"))

    quiet = traits.Bool(
        mandatory=False, argstr="-quiet",
        desc="Don't display output during operation")


class MRMathOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='The resultant image')


class MRMath(CommandLine):
    """
    Extracts the gradient information in MRtrix format from a DWI image
    """
    _cmd = 'mrmath'
    input_spec = MRMathInputSpec
    output_spec = MRMathOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = self._gen_outfilename()
        else:
            assert False
        return fname

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            filename = self.inputs.out_file
        else:
            base, ext = split_extension(
                os.path.basename(self.inputs.in_files[0]))
            filename = os.path.join(
                os.getcwd(),
                "{}_{}{}".format(base, self.inputs.operation, ext))
        return filename


# =============================================================================
# MR math
# =============================================================================

class MRCalcInputSpec(CommandLineInputSpec):

    operands = traits.List(
        traits.Any(), argstr='%s',
        mandatory=True, position=-3,
        desc="Diffusion weighted images with graident info")

    out_file = File(genfile=True, argstr='%s', position=-1,
                    desc="Extracted DW or b-zero images")
    out_ext = traits.Str(desc='Extention of the output file.')

    operation = traits.Enum(
        'abs', 'neg', 'sqrt', 'exp', 'log', 'log10', 'cos', 'sin', 'tan',
        'cosh', 'sinh', 'tanh', 'acos', 'asin', 'atan', 'acosh', 'asinh',
        'atanh', 'round', 'ceil', 'floor', 'isnan', 'isinf', 'finite', 'real',
        'imag', 'phase', 'conj', 'add', 'subtract', 'multiply', 'divide',
        'pow', 'min', 'max', 'lt', 'gt', 'le', 'ge', 'eq', 'neq', 'complex',
        'if', mandatory=True, argstr='-%s',
        position=-2, desc=("Operation to apply to the files"))

    quiet = traits.Bool(
        mandatory=False, argstr="-quiet",
        desc="Don't display output during operation")


class MRCalcOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='The resultant image')


class MRCalc(CommandLine):
    """
    Performs a mathematical calculation on the given images
    """
    _cmd = 'mrcalc'
    input_spec = MRCalcInputSpec
    output_spec = MRCalcOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = self._gen_outfilename()
        else:
            assert False
        return fname

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            filename = self.inputs.out_file
        else:
            if isdefined(self.inputs.out_ext):
                ext = self.inputs.out_ext
            else:
                _, ext = split_extension(
                    os.path.basename(self.inputs.operands[0]))
            filename = os.getcwd()
            for op in self.inputs.operands:
                try:
                    op_str = split_extension(os.path.basename(op))[0]
                except:
                    op_str = str(float(op))
                filename += '_' + op_str
            filename += '_' + self.inputs.operation + ext
        return filename


class ExtractFSLGradientsInputSpec(CommandLineInputSpec):
    in_file = traits.Either(
        File, Directory, exists=True, argstr='%s', mandatory=True, position=0,
        desc="Diffusion weighted images with graident info")
    out_file = traits.File(
        'out.mif', usedefault=True, argstr='%s', position=-1,
        desc="Dummy output file, which isn't used.")
    bvecs_file = File(genfile=True, argstr='-export_grad_fsl %s', position=1,
                      desc=("Extracted gradient encoding directions in FSL "
                            "format"))
    bvals_file = File(genfile=True, argstr='%s', position=2,
                      desc=("Extracted graident encoding b-values in FSL "
                            "format"))


class ExtractFSLGradientsOutputSpec(TraitedSpec):
    bvecs_file = File(exists=True,
                      desc='Extracted encoding gradient directions')
    bvals_file = File(exists=True,
                      desc='Extracted encoding gradient b-values')


class ExtractFSLGradients(CommandLine):
    """
    Extracts the gradient information in MRtrix format from a DWI image
    """
    _cmd = 'mrconvert'
    input_spec = ExtractFSLGradientsInputSpec
    output_spec = ExtractFSLGradientsOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['bvecs_file'] = self._gen_grad_filename('bvec')
        outputs['bvals_file'] = self._gen_grad_filename('bval')
        return outputs

    def _gen_filename(self, name):
        if name == 'bvecs_file':
            fname = self._gen_grad_filename('bvec')
        elif name == 'bvals_file':
            fname = self._gen_grad_filename('bval')
        else:
            assert False
        return fname

    def _gen_grad_filename(self, comp):
        filename = getattr(self.inputs, comp + 's_file')
        if not isdefined(filename):
            base, _ = split_extension(os.path.basename(self.inputs.in_file))
            filename = os.path.join(
                os.getcwd(), "{base}_{comp}s.{comp}".format(base=base,
                                                            comp=comp))
        return filename


# =============================================================================
# Extract b0 or DW images
# =============================================================================

class ExtractDWIorB0InputSpec(CommandLineInputSpec):
    in_file = traits.Either(File, Directory, exists=True, argstr='%s',
                            mandatory=True, position=0,
                            desc="Diffusion weighted images with graident info")

    out_file = File(genfile=True, argstr='%s', position=-1,
                    desc="Extracted DW or b-zero images")

    bzero = traits.Bool(argstr='-bzero', position=1,  # @UndefinedVariable
                        desc="Extract b-zero images instead of DDW images")

    quiet = traits.Bool(
        mandatory=False, argstr="-quiet",
        desc="Don't display output during operation")

    out_ext = traits.Str(desc='Extention of the output file.')

    grad = traits.Str(
        mandatory=False, argstr='-grad %s',
        desc=("specify the diffusion-weighted gradient scheme used in the  "
              "acquisition. The program will normally attempt to use the  "
              "encoding stored in the image header. This should be supplied  "
              "as a 4xN text file with each line is in the format [ X Y Z b ],"
              " where [ X Y Z ] describe the direction of the applied  "
              "gradient, and b gives the b-value in units of s/mm^2."))

    fslgrad = traits.Tuple(
        File(exists=True, desc="gradient directions file (bvec)"),  # @UndefinedVariable @IgnorePep8
        File(exists=True, desc="b-values (bval)"),  # @UndefinedVariable @IgnorePep8
        argstr='-fslgrad %s %s', mandatory=False,
        desc=("specify the diffusion-weighted gradient scheme used in the "
              "acquisition in FSL bvecs/bvals format."))


class ExtractDWIorB0OutputSpec(TraitedSpec):

    out_file = File(exists=True, desc='Extracted DW or b-zero images')


class ExtractDWIorB0(CommandLine):
    """
    Extracts the gradient information in MRtrix format from a DWI image
    """
    _cmd = 'dwiextract'
    input_spec = ExtractDWIorB0InputSpec
    output_spec = ExtractDWIorB0OutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = self._gen_outfilename()
        else:
            assert False
        return fname

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            filename = self.inputs.out_file
        else:
            if isdefined(self.inputs.out_ext):
                ext = self.inputs.out_ext
                base, _ = split_extension(os.path.basename(
                    self.inputs.in_file))
            else:
                base, ext = split_extension(os.path.basename(
                    self.inputs.in_file))
            if isdefined(self.inputs.bzero):
                suffix = 'b0'
            else:
                suffix = 'dw'
            filename = os.path.join(
                os.getcwd(), "{}_{}{}".format(base, suffix, ext))
        return filename



from nipype.interfaces.afni.base import AFNICommand, AFNICommandInputSpec
from nipype.interfaces.base import File, traits, TraitedSpec


class TprojectInputSpec(AFNICommandInputSpec):

    in_file = File(desc='input file to 3dTproject', argstr='-input %s',
                   exists=True, mandatory=True)
    out_file = File(name_template="%s_filt", desc='output image file name',
                    argstr='-prefix %s', name_source="in_file")
    ort = File(desc='remove each column in file. Multiple ort are allowed',
               argstr='-ort %s', exists=True)
    polort = traits.Int(desc='remove polynomials up to degree specified',
                        argstr='-polort %s')
    dsort = File(
        desc='Remove the 3D+time time series in the specified dataset',
        argstr='-dsort %s', exists=True)
    passband = traits.Tuple(
        (traits.Float(), traits.Float()), argstr='-passband %f %f',
        desc='Remove all frequencies EXCEPT those in the range provided.')
    stopband = traits.Tuple(
        (traits.Float(), traits.Float()), argstr='-stopband %f %f',
        desc='Remove all frequencies in the range provided.')
    delta_t = traits.Float(desc='time step for frequency calculations',
                           argstr='-dt %f')
    mask = File(desc='Only operate on voxels nonzero in the provided mask.',
                argstr='-mask %s', exists=True, xor=['automask'])
    automask = traits.Bool(desc='the program will generate the mask to use for'
                           'calculations.', argstr='-automask', xor=['mask'])
    blur = traits.Float(desc='Blur (inside the mask only) with a filter that'
                        'has width (FWHM) of fff millimeters.',
                        argstr='-blur %f')
    normalize = traits.Bool(desc='Normalize each output time series to have'
                            'sum of squares = 1. This is the LAST operation.',
                            argstr='-norm')


class TprojectOutputSpec(TraitedSpec):
    out_file = File(desc='filtered file', exists=True)


class Tproject(AFNICommand):
    """This program projects (detrends) out various 'nuisance' time series from
    each voxel in the input dataset.  Note that all the projections are done
    via linear regression, including the frequency-based parameters such as
    '-passband'.  In this way, you can bandpass time-censored data, and at the
    same time, remove other time series of no interest (e.g., physiological
    estimates, motion parameters).

    Examples
    ========
    >>> banana.interfaces.afni import Tproject
    >>> TP = Tproject()
    >>> TP.inputs.in_file = 'functional.nii'
    >>> TP.inputs.passband = (0.01, 0.08)
    >>> TP.delta_t = 0.754
    >>> TP.inputs.outputtype = "NIFTI"
    >>> res = TP.run()
    """

    _cmd = '3dTproject'
    input_spec = TprojectInputSpec
    output_spec = TprojectOutputSpec



from .fibre_est import EstimateFOD, AverageResponse
from .preproc import (
    DWIPreproc, DWI2Mask, DWIBiasCorrect, DWIDenoise,
    DWIIntensityNorm)
from .utils import (
    MRConvert, MRCat, MRCrop, MRPad, MRMath, MRCalc, ExtractFSLGradients,
    ExtractDWIorB0)
from .tracking import GlobalTractography



import os
import os.path as op
from copy import deepcopy
from collections import defaultdict
import logging
import re
import numpy as np
import nibabel as nib
from nipype.interfaces.base import (
    TraitedSpec, traits, BaseInterface, BaseInterfaceInputSpec, File,
    Directory, isdefined)
from banana.exceptions import BananaUsageError
logger = logging.getLogger('banana')


class ToPolarCoordsInputSpec(BaseInterfaceInputSpec):
    in_dir = Directory(exists=True, mandatory=True)
    in_fname_re = traits.Str(
        r'.*_(?P<channel>\d+)_(?P<echo>\d+)_(?P<axis>[A-Z]+)\.nii\.gz',
        usedefault=True, desc=(
            "Regex to extract the channel, echo and axis "
            "(i.e. real or imaginary) information from the input file name. "
            "Must incluce named groups for 'channel', 'echo' and 'axis'"))
    out_fname_str = traits.Str(
        'coil_{channel}_{echo}.nii.gz', usedefault=True,
        desc=("The format string used to generate the save channel filenames. "
              "Must use the 'channel' and 'echo' field names"))
    real_label = traits.Str('REAL', usedefault=True, desc=(
        "The label used to specify the real component image"))
    imaginary_label = traits.Str('IMAGINARY', usedefault=True, desc=(
        "The label used to specify the real component image"))
    combined_dir = Directory(genfile=True, desc=(
        "Output directory for coil magnitude and phase images. "
        "Files will be saved with the name "
        "'Raw_Coil_<channel>_<echo>.nii.gz'"))
    magnitudes_dir = Directory(genfile=True, desc=(
        "Output directory for coil magnitude images."))
    phases_dir = Directory(genfile=True, desc=(
        "Output directory for coil phase images"))


class ToPolarCoordsOutputSpec(TraitedSpec):
    combined_images = traits.List(
        File(exists=True),
        desc="List of combined images for each echo using least squares")
    first_echo = File(exists=True,
                      desc="The first echo of the combined images")
    last_echo = File(exists=True,
                     desc="The last echo of the combined images")
    coil_magnitudes = traits.List(
        traits.List(File(exists=True)),
        desc=("List of magnitude images for each coil for each echo"))
    coil_phases = traits.List(
        traits.List(File(exists=True)),
        desc=("List of magnitude images for each coil for each echo"))
    combined_dir = Directory(exists=True, desc=(
        "Output directory for combined magnitude images for each echo time "))
    magnitudes_dir = Directory(exists=True, desc=(
        "Output directory for coil magnitude images"))
    phases_dir = Directory(exists=True, desc=(
        "Output directory for coil phase images"))


class ToPolarCoords(BaseInterface):
    """
    Takes all REAL and IMAGINARY pairs in current directory and prepares
    them for Phase and QSM processing.

    1. Existence of pairs is checked
    2. Files are load/save cycled for formatting and rename for consistency
    3. Magnitude and Phase components are produced
    4. Coils are combined for single magnitude images per echo
    """
    input_spec = ToPolarCoordsInputSpec
    output_spec = ToPolarCoordsOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):  # @UnusedVariable
        print("in-dir: " + self.inputs.in_dir)
        outputs = self._outputs().get()
        # Get names for output directories
        combined_dir = outputs['combined_dir'] = self._gen_filename(
            'combined_dir')
        mags_dir = outputs['magnitudes_dir'] = self._gen_filename(
            'magnitudes_dir')
        phases_dir = outputs['phases_dir'] = self._gen_filename(
            'phases_dir')
        # Ensure output directories exist
        os.makedirs(combined_dir, exist_ok=True)
        os.makedirs(mags_dir, exist_ok=True)
        os.makedirs(phases_dir, exist_ok=True)
        outputs['combined_images'] = []
        coil_mags = outputs['coil_magnitudes'] = []
        coil_phases = outputs['coil_phases'] = []
        # A default dict with three levels of keys to hold the file names
        # sorted into echo, channel and complex axis
        paths = defaultdict(lambda: defaultdict(dict))
        # Compile regular expression for extracting channel, echo and
        # complex axis indices from input file names
        fname_re = re.compile(self.inputs.in_fname_re)
        for fname in os.listdir(self.inputs.in_dir):
            match = fname_re.match(fname)
            if match is None:
                logger.warning("Skipping '{}' file in '{}' as it doesn't "
                               "match expected filename pattern for raw "
                               "channel files ('{}')"
                               .format(fname, self.inputs.in_dir,
                                       self.inputs.in_fname_re))
                continue
            paths[match.group('echo')][match.group('channel')][
                match.group('axis')] = op.join(self.inputs.in_dir, fname)

        first_echo_index = min(paths.keys())
        last_echo_index = max(paths.keys())

        for echo_i, channels in paths.items():
            # Variables to hold combined coil images
            combined_array = None
            normaliser_array = None
            echo_coil_mags = []
            echo_coil_phases = []
            for channel_i, axes in channels.items():
                # Load image real and imaginary data and remove extreme values
                img_arrays = {}
                for ax, fname in axes.items():
                    img = nib.load(fname)
                    img_array = img.get_fdata()
                    # Replace extreme values with random value
                    img_array[img_array == 2048] = 0.02 * np.random.rand()
                    img_arrays[ax] = img_array

                # Calculate magnitude and phase from coil data
                cmplx = (img_arrays[self.inputs.real_label] +
                         img_arrays[self.inputs.imaginary_label] * 1j)

                # Calculate and save magnitude image
                mag_array = np.abs(cmplx)
                mag_img = nib.Nifti1Image(mag_array, img.affine, img.header)
                mag_path = op.join(
                    mags_dir,
                    self.inputs.out_fname_str.format(channel=channel_i,
                                                     echo=echo_i))
                echo_coil_mags.append(mag_path)
                nib.save(mag_img, mag_path)

                # Save phase image
                phase_array = np.angle(cmplx)
                phase_img = nib.Nifti1Image(phase_array, img.affine,
                                            img.header)
                phase_path = op.join(
                    phases_dir,
                    self.inputs.out_fname_str.format(channel=channel_i,
                                                     echo=echo_i))
                echo_coil_phases.append(phase_path)
                nib.save(phase_img, phase_path)

                # Add coil data to combined coil data
                if combined_array is None:
                    combined_array = deepcopy(mag_array) ** 2
                    normaliser_array = deepcopy(mag_array)
                else:
                    combined_array += mag_array ** 2
                    normaliser_array += mag_array
            coil_mags.append(echo_coil_mags)
            coil_phases.append(echo_coil_phases)
            # Normalise combined sum of squares image, save and append
            # to list of combined echoes
            combined_array /= normaliser_array
            combined_array[np.isnan(combined_array)] = 0
            # Generate filename and append ot list of combined coil images
            combined_fname = op.join(combined_dir,
                                     'echo_{}.nii.gz'.format(echo_i))
            combined_img = nib.Nifti1Image(combined_array, img.affine,
                                           img.header)
            nib.save(combined_img, combined_fname)
            outputs['combined_images'].append(combined_fname)
            if echo_i == first_echo_index:
                outputs['first_echo'] = combined_fname
            if echo_i == last_echo_index:
                outputs['last_echo'] = combined_fname
        return outputs

    def _gen_filename(self, name):
        if name == 'combined_dir':
            fname = op.abspath(self.inputs.combined_dir
                               if isdefined(self.inputs.combined_dir)
                               else 'combined_images')
        elif name == 'magnitudes_dir':
            fname = op.abspath(self.inputs.magnitudes_dir
                               if isdefined(self.inputs.magnitudes_dir)
                               else 'magnitudes_dir')
        elif name == 'phases_dir':
            fname = op.abspath(self.inputs.phases_dir
                               if isdefined(self.inputs.phases_dir)
                               else 'phases_dir')
        else:
            assert False
        return fname


class HIPCombineChannelsInputSpec(BaseInterfaceInputSpec):

    magnitudes_dir = Directory(exists=True, desc=(
        "Input directory containing coil magnitude images."))
    phases_dir = Directory(exists=True, desc=(
        "Input directory containing coil phase images."))
    in_fname_re = traits.Str(
        'coil_(?P<channel>\d+)_(?P<echo>\d+)\.nii\.gz', usedefault=True,
        desc=("The format string used to generate the save channel filenames. "
              "Must use the 'channel' and 'echo' field names"))
    magnitude = File(genfile=True, desc="Combined magnitude image")
    phase = File(genfile=True, desc="Combined phase image")
    q = File(genfile=True, desc="Q image")


class HIPCombineChannelsOutputSpec(TraitedSpec):

    magnitude = File(exists=True, desc="Combined magnitude image")
    phase = File(exists=True, desc="Combined phase image")
    q = File(exists=True, desc="Q image")


class HIPCombineChannels(BaseInterface):
    """
    Apply Laplacian unwrapping from STI suite to each coil
    """
    input_spec = HIPCombineChannelsInputSpec
    output_spec = HIPCombineChannelsOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):  # @UnusedVariable
        outputs = self._outputs().get()
        mag_fname = outputs['magnitude'] = self._gen_filename('magnitude')
        phase_fname = outputs['phase'] = self._gen_filename('phase')
        q_fname = outputs['q'] = self._gen_filename('q')
        mag_paths = defaultdict(dict)
        phase_paths = defaultdict(dict)
        # Compile regular expression for extracting channel, echo and
        # complex axis indices from input file names
        fname_re = re.compile(self.inputs.in_fname_re)
        for dpath, dct in ((self.inputs.magnitudes_dir, mag_paths),
                           (self.inputs.phases_dir, phase_paths)):
            for fname in os.listdir(dpath):
                match = fname_re.match(fname)
                if match is None:
                    logger.warning("Skipping '{}' file in '{}' as it doesn't "
                                   "match expected filename pattern for raw "
                                   "channel files ('{}')"
                                   .format(fname, dpath,
                                           self.inputs.in_fname_re))
                    continue
            dct[match.group('channel')][match.group('echo')] = op.join(dpath,
                                                                       fname)
        if len(mag_paths) != len(phase_paths):
            raise BananaUsageError(
                "Mismatching number of channels between magnitude and phase "
                "channels")
        hip = None
        for chann_i in mag_paths:
            if len(mag_paths[chann_i]) != 2:
                raise BananaUsageError(
                    "Expected exactly two echos for channel magnitude {}, "
                    "found {}".format(chann_i, len(mag_paths[chann_i])))
            if len(phase_paths[chann_i]) != 2:
                raise BananaUsageError(
                    "Expected exactly two echos for channel magnitude {}, "
                    "found {}".format(chann_i, len(phase_paths[chann_i])))
            mag1 = nib.load(mag_paths[chann_i][0])
            phase1 = nib.load(phase_paths[chann_i][0])
            mag2 = nib.load(mag_paths[chann_i][1])
            phase2 = nib.load(phase_paths[chann_i][1])

            # Get array data
            mag1_array = mag1.get_fdata()
            phase1_array = phase1.get_fdata()
            mag2_array = mag2.get_fdata()
            phase2_array = phase2.get_fdata()

            if hip is None:
                hip = np.zeros(mag1_array.shape)
                sum_mag = np.zeros(mag1_array.shape)
            hip += mag1_array * mag2_array * np.exp(
                -1j * (phase1_array - phase2_array))
            sum_mag += mag1_array * mag2_array
        # Get magnitude and phase
        phase = np.angle(hip)
        mag = np.abs(hip)
        q = mag / sum_mag
        # Create NIfTI images
        phase_img = nib.Nifti1Image(phase, phase1.affine, phase1.header)
        mag_img = nib.Nifti1Image(mag, mag1.affine, mag1.header)
        q_img = nib.Nifti1Image(q, mag1.affine, mag1.header)
        # Save NIfTIs
        nib.save(phase_img, phase_fname)
        nib.save(mag_img, mag_fname)
        nib.save(q_img, q_fname)
        return outputs

    def _gen_filename(self, name):
        if name == 'magnitude':
            fname = op.abspath(self.inputs.magnitude
                               if isdefined(self.inputs.magnitude)
                               else 'magnitude')
        elif name == 'phase':
            fname = op.abspath(self.inputs.phase
                               if isdefined(self.inputs.phase) else 'phase')
        elif name == 'q':
            fname = op.abspath(self.inputs.q if isdefined(self.inputs.q)
                               else 'q')
        else:
            assert False
        return fname



import os
import os.path as op
import logging
import tempfile
import shutil
from copy import copy
from pprint import pformat
from itertools import chain
from unittest import TestCase
from arcana.exceptions import ArcanaNameError
from arcana import (InputFilesets, InputFields, BasicRepo, XnatRepo, SingleProc,
                    Field, Fileset, ModulesEnv, StaticEnv)
from arcana.data.spec import BaseInputSpec
from arcana.exceptions import (
    ArcanaInputMissingMatchError, ArcanaMissingDataException)
from banana.exceptions import BananaTestSetupError, BananaUsageError
import banana.file_format  # @UnusedImport


logger = logging.getLogger('arcana')
logger.setLevel(logging.WARNING)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(levelname)s - %(message)s")
handler.setFormatter(formatter)

# logger = logging.getLogger('nipype.workflow')
# logger.setLevel(logging.DEBUG)
# handler = logging.StreamHandler()
# formatter = logging.Formatter("%(levelname)s - %(message)s")
# handler.setFormatter(formatter)

logger.addHandler(handler)

logging.getLogger("urllib3").setLevel(logging.WARNING)

try:
    TEST_DIR = os.environ['BANANA_TEST_DIR']
except KeyError:
    TEST_DIR = tempfile.mkdtemp()

USE_MODULES = 'BANANA_TEST_USE_MODULES' in os.environ


TEST_CACHE_DIR = op.join(TEST_DIR, 'cache')


class PipelineTester(TestCase):
    """
    Runs pipelines within a Study class and compares the results
    against reference data from previous runs

    Class attributes
    ----------------
    name : str
        A unique name (among other unittests to be run in same batch) for the
        test class
    study_class : type(Study)
        The study class run tests against
    ref_repo : Repository
        The repository to draw the reference data from
    parameters : dict[str, str | float | int | datetime]
        The parameters passed to the study when it is initialised
    warn_missing_tests : bool
        Whether to raise a warning if there is a pipeline in the study class
        that does not have a test in tester class
    """

    parameters = {}
    warn_missing_tests = True
    environment = ModulesEnv() if USE_MODULES else StaticEnv()

    def setUp(self):
        if not hasattr(self, 'study_class'):
            raise BananaTestSetupError(
                "{} must have a 'study_class' attribute corresponding to the "
                "Study class to test".format(type(self).__name__))
        if not hasattr(self, 'ref_repo'):
            raise BananaTestSetupError(
                "{} must have a 'ref_repo' attribute corresponding to the "
                "Study class to test".format(type(self).__name__))
        base_test_dir = op.join(TEST_DIR, self.name)
        self.work_dir = op.join(base_test_dir, 'work')
        repo_root = op.join(base_test_dir, 'repo')
        os.makedirs(self.work_dir, exist_ok=True)
        os.makedirs(repo_root, exist_ok=True)
        # Create repository to hold outputs
        self.output_repo = BasicRepo(repo_root, depth=2)
        # Create inputs for reference study
        self.inputs = {}
        for spec in self.study_class.data_specs():
            # Create an input for each entry in the class specificiation
            if spec.is_fileset:
                inpt = InputFilesets(
                    spec.name, spec.name, repository=self.ref_repo)
            else:
                inpt = InputFields(
                    spec.name, spec.name, dtype=spec.dtype,
                    repository=self.ref_repo)
            # Check whether a corresponding data exists in the reference repo
            try:
                inpt.match(self.ref_repo.cached_tree(),
                           valid_formats=getattr(spec, 'valid_formats', None))
            except ArcanaInputMissingMatchError:
                continue
            self.inputs[spec.name] = inpt
        # Create the reference study
        self.ref_study = self.study_class(
            self.name,
            repository=self.ref_repo,
            processor=self.work_dir,
            inputs=self.inputs.values(),
            environment=self.environment,
            parameters=self.parameters)
        # Get set of all pipelines to test
        if self.warn_missing_tests:
            pipelines_to_test = set(
                s.pipeline_getter for s in self.study_class.data_specs()
                if s.derived)
            test_names = [m for m in dir(self) if m.startswith('test_')]
            for pipeline_getter in pipelines_to_test:
                expected_test_name = 'test_' + pipeline_getter
                if expected_test_name not in test_names:
                    logger.warning("Did not find test for '{}' pipeline in {}"
                                   "tester. Add to 'suppress_warning' class "
                                   "attr to suppress this message".format(
                                       pipeline_getter, self.name))

    def run_pipeline_test(self, pipeline_getter, add_inputs=(),
                          test_criteria=None, pipeline_args=None):
        """
        Runs a pipeline and tests its outputs against the reference data

        Parameters
        ----------
        pipeline_getter : str
            The name of the pipeline to test
        add_inputs : list[str]
            Inputs that are required in the output study for the pipeline to
            run, in addition to the direct inputs of the pipeline, i.e. ones
            that are tested for with the 'provided' method in the pipeline
            construction
        test_criteria : dct[str, *] | None
            A dictionary containing the criteria by which to determine equality
            for the derived filesets. The keys are spec-names and the values
            are specific to the format of the fileset. If a spec-name is not
            in the dictionary or None is provided then the default
            criteria are used for each fileset test.
        """
        if test_criteria is None:
            test_criteria = {}
        if pipeline_args is None:
            pipeline_args = {}
        # A study with all inputs provided to determine which inputs are needed
        # by the pipeline
        ref_pipeline = self.ref_study.pipeline(pipeline_getter,
                                               pipeline_args=pipeline_args)
        inputs = []
        for spec_name in chain(ref_pipeline.input_names, add_inputs):
            try:
                inputs.append(self.inputs[spec_name])
            except KeyError:
                pass  # Inputs with a default value
        # Set up output study
        output_study = self.study_class(
            pipeline_getter,
            repository=self.output_repo,
            processor=SingleProc(self.work_dir, reprocess='force'),
            environment=self.environment,
            inputs=inputs,
            parameters=self.parameters,
            subject_ids=self.ref_study.subject_ids,
            visit_ids=self.ref_study.visit_ids,
            enforce_inputs=False,
            fill_tree=True)
        for spec_name in ref_pipeline.output_names:
            for ref, test in zip(self.ref_study.data(spec_name),
                                 output_study.data(spec_name)):
                if ref.is_fileset:
                    try:
                        self.assertTrue(
                            test.contents_equal(
                                ref, **test_criteria.get(spec_name, {})),
                            "'{}' fileset generated by {} in {} doesn't match "
                            "reference".format(spec_name, pipeline_getter,
                                               self.study_class))
                    except Exception:
                        if hasattr(test, 'headers_diff'):
                            header_diff = test.headers_diff(ref)
                            if header_diff:
                                print("Headers don't match on {}"
                                      .format(header_diff))
                                print("Test header:\n{}".format(
                                    pformat(test.get_header())))
                                print("Reference header:\n{}".format(
                                    pformat(ref.get_header())))
                            else:
                                print("Image RMS diff: {}"
                                      .format(test.rms_diff(ref)))
                                test.contents_equal(ref)
                        raise
                else:
                    self.assertEqual(
                        test.value, ref.value,
                        "value for {} ({}) generated by {} in {} doesn't "
                        "match reference ({})".format(
                            spec_name, test.value, pipeline_getter,
                            self.study_class, ref.value))

    @classmethod
    def generate_test_data(cls, study_class, in_repo, out_repo,
                           in_server=None, out_server=None, work_dir=None,
                           parameters=(), include=None, skip=(),
                           include_bases=(), reprocess=False, repo_depth=0,
                           modules_env=False, clean_work_dir=True,
                           loggers=('nipype.workflow', 'arcana', 'banana')):
        """
        Generates reference data for a pipeline tester unittests given a study
        class and set of parameters

        Parameters
        ----------
        study_class : type(Study)
            The path to the study class to test, e.g. banana.study.MriStudy
        in_repo : str
            The path to repository that houses the input data
        out_repo : str
            If the 'xnat_server' argument is provided then out
            is interpreted as the project ID to use the XNAT
            server (the project must exist already). Otherwise
            it is interpreted as the path to a basic repository
        in_server : str | None
            The server to download the input data from
        out_server : str | None
            The server to upload the reference data to
        work_dir : str
            The work directory
        parameters : dict[str, *]
            Parameter to set when initialising the study
        include : list[str] | None
            Spec names to include in the output repository. If None all names
            except those listed in 'skip' are included
        skip : list[str]
            Spec names to skip in the generation process. Only valid if
            'include' is None
        include_bases : list[type(Study)]
            List of base classes in which all entries in their data
            specification are added to the list to include
        reprocess : bool
            Whether to reprocess the generated datasets
        repo_depth : int
            The depth of the input repository
        modules_env : bool
            Whether to use modules environment or not
        clean_work_dir : bool
            Whether to clean the Nipype work directory or not
        """

        for logger_name in loggers:
            logger = logging.getLogger(logger_name)
            logger.setLevel(logging.INFO)
            handler = logging.StreamHandler()
            formatter = logging.Formatter("%(levelname)s - %(message)s")
            handler.setFormatter(formatter)
            logger.addHandler(handler)

        if work_dir is None:
            work_dir = tempfile.mkdtemp()
        else:
            work_dir = work_dir

        if study_class.__name__.endswith('Study'):
            study_name = study_class.__name__[:-len('Study')]
        else:
            study_name = study_class.__name__

        # Get output repository to write the data to
        if in_server is not None:
            in_repo = XnatRepo(project_id=in_repo, server=in_server,
                               cache_dir=op.join(work_dir, 'xnat-cache'))
        else:
            in_repo = BasicRepo(in_repo, depth=repo_depth)

        temp_repo_root = op.join(work_dir, 'temp-repo')
        if os.path.exists(temp_repo_root) and reprocess:
            shutil.rmtree(temp_repo_root)
        os.makedirs(temp_repo_root, exist_ok=True)

        temp_repo = BasicRepo(temp_repo_root, depth=repo_depth)

        inputs = None
        for session in in_repo.tree().sessions:
            session_inputs = []
            for item in chain(session.filesets, session.fields):
                if isinstance(item, Fileset):
                    inpt = InputFilesets(item.basename, item.basename,
                                         item.format, repository=in_repo)
                else:
                    inpt = InputFields(item.name, item.name, item.dtype,
                                       repository=in_repo)
                try:
                    spec = study_class.data_spec(inpt)
                except ArcanaNameError:
                    print("Skipping {} as it doesn't match a spec in {}"
                          .format(item, study_class))
                else:
                    session_inputs.append(inpt)
            session_inputs = sorted(session_inputs)
            if inputs is not None and session_inputs != inputs:
                raise BananaUsageError(
                    "Inconsistent inputs ({} and {}) found in sessions of {}"
                    .format(inputs, session_inputs, in_repo))
            else:
                inputs = session_inputs

        if modules_env:
            env = ModulesEnv()
        else:
            env = StaticEnv()

        study = study_class(
            study_name,
            repository=temp_repo,
            processor=SingleProc(
                work_dir, reprocess=reprocess,
                clean_work_dir_between_runs=clean_work_dir,
                prov_ignore=(
                    SingleProc.DEFAULT_PROV_IGNORE +
                    ['.*/pkg_version',
                     'workflow/nodes/.*/requirements/.*'])),
            environment=env,
            inputs=inputs,
            parameters=parameters,
            subject_ids=in_repo.tree().subject_ids,
            visit_ids=in_repo.tree().visit_ids,
            fill_tree=True)

        if include is None:
            # Get set of methods that could override pipeline getters in
            # base classes that are not included
            potentially_overridden = set()
            for cls in chain(include_bases, [study_class]):
                potentially_overridden.update(cls.__dict__.keys())

            include = set()
            for base in study_class.__mro__:
                if not hasattr(base, 'add_data_specs'):
                    continue
                for spec in base.add_data_specs:
                    if isinstance(spec, BaseInputSpec) or spec.name in skip:
                        continue
                    if (base is study_class or base in include_bases or
                            spec.pipeline_getter in potentially_overridden):
                        include.add(spec.name)

        # Generate all derived data
        for spec_name in sorted(include):
            study.data(spec_name)

        # Get output repository to write the data to
        if out_server is not None:
            out_repo = XnatRepo(project_id=out_repo, server=out_server,
                                cache_dir=op.join(work_dir, 'xnat-cache'))
        else:
            out_repo = BasicRepo(out_repo, depth=repo_depth)

        # Upload data to repository
        for spec in study.data_specs():
            try:
                data = study.data(spec.name, generate=False)
            except ArcanaMissingDataException:
                continue
            for item in data:
                if not item.exists:
                    logger.info("Skipping upload of non-existant {}"
                                .format(item.name))
                    continue
                if skip is not None and item.name in skip:
                    logger.info("Forced skip of {}".format(item.name))
                    continue
                if item.is_fileset:
                    item_cpy = Fileset(
                        name=item.name, format=item.format,
                        frequency=item.frequency, path=item.path,
                        aux_files=copy(item.aux_files),
                        subject_id=item.subject_id, visit_id=item.visit_id,
                        repository=out_repo, exists=True)
                else:
                    item_cpy = Field(
                        name=item.name, value=item.value, dtype=item.dtype,
                        frequency=item.frequency, array=item.array,
                        subject_id=item.subject_id, visit_id=item.visit_id,
                        repository=out_repo, exists=True)
                logger.info("Uploading {}".format(item_cpy))
                item_cpy.put()
                logger.info("Uploaded {}".format(item_cpy))
        logger.info("Finished generating and uploading test data for {}"
                    .format(study_class))


if __name__ == '__main__':
    from banana.study.mri.base import MriStudy
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('generate', nargs='+', help="datasets to generate")
    parser.add_argument('--data_dir', default=op.expanduser('~/Data'),
                        help="The data dir to use")
    parser.add_argument('--dont_clean_work_dir', action='store_true',
                        default=False, help="Don't clean work dir")
    args = parser.parse_args()

    print("Generating test data for {}".format(args.generate))

    if 'mri' in args.generate:

        PipelineTester.generate_test_data(
            MriStudy, op.join(args.data_dir, 'mri'), 'TESTBANANAMRI',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 'mri-work'),
            reprocess=False, repo_depth=0, modules_env=True,
            skip=['channels', 'mag_channels', 'phase_channels'],
            clean_work_dir=(not args.dont_clean_work_dir),
            parameters={'mni_template_resolution': 1})

    if 'mri2' in args.generate:

        PipelineTester.generate_test_data(
            MriStudy, op.join(args.data_dir, 'mri2'), 'TESTBANANAMRI2',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 'mri2-work'),
            reprocess=False, repo_depth=0, modules_env=True,
            clean_work_dir=(not args.dont_clean_work_dir),
            include=['brain_coreg'],
            skip=['template', 'template_brain', 'template_mask'],
            parameters={
                'coreg_method': 'flirt'})

    if 'base3' in args.generate:

        PipelineTester.generate_test_data(
            MriStudy, op.join(args.data_dir, 'mri'), 'TESTBANANAMRI3',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 'mri3-work'),
            reprocess=False, repo_depth=0, modules_env=True,
            clean_work_dir=(not args.dont_clean_work_dir),
            include=['brain_coreg'])

    if 'bold' in args.generate:
        from banana.study.mri.bold import BoldStudy

        PipelineTester.generate_test_data(
            BoldStudy, op.join(args.data_dir, 'bold'), 'TESTBANANABOLD',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 'bold-work'),
            reprocess=False, repo_depth=0, modules_env=True,
            clean_work_dir=(not args.dont_clean_work_dir),
            skip=['field_map_delta_te', 'cleaned_file'],
            parameters={
                'mni_template_resolution': 2})

    if 't1' in args.generate:
        from banana.study.mri.t1 import T1Study

        PipelineTester.generate_test_data(
            T1Study, op.join(args.data_dir, 't1'), 'TESTBANANAT1',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 't1-work'),
            skip=['t2_coreg'],
            include=None,
            reprocess=False, repo_depth=1, modules_env=True,
            clean_work_dir=(not args.dont_clean_work_dir))

    if 't2' in args.generate:
        from banana.study.mri.t2 import T2Study

        PipelineTester.generate_test_data(
            T2Study, op.join(args.data_dir, 't2'), 'TESTBANANAT2',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 't2-work'),
            reprocess=False, repo_depth=0, modules_env=True,
            clean_work_dir=(not args.dont_clean_work_dir))

    if 't2star' in args.generate:
        from banana.study.mri.t2star import T2starStudy

        PipelineTester.generate_test_data(
            T2starStudy, op.join(args.data_dir, 't2star'), 'TESTBANANAT2S',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 't2star-work'),
            reprocess=False, repo_depth=0, modules_env=True,
            parameters={
                'mni_template_resolution': 2},
            clean_work_dir=(not args.dont_clean_work_dir))

    if 'dwi' in args.generate:
        from banana.study.mri.dwi import DwiStudy
        from banana.study.mri.epi import EpiSeriesStudy

        PipelineTester.generate_test_data(
            DwiStudy, op.join(args.data_dir, 'dwi'), 'TESTBANANADWI',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 'dwi-work'),
            skip=['coreg_ref_wmseg', 'field_map_mag', 'brain_coreg',
                  'field_map_phase', 'moco', 'align_mats', 'moco_par',
                  'field_map_delta_te', 'norm_intensity', 'brain_mask_coreg',
                  'norm_intens_fa_template', 'norm_intens_wm_mask',
                  'connectome'],
            include_bases=[EpiSeriesStudy],
            parameters={
                'num_global_tracks': int(1e6)}, include=None,
            reprocess=False, repo_depth=1, modules_env=True,
            clean_work_dir=(not args.dont_clean_work_dir))

    if 'dwi2' in args.generate:
        from banana.study.mri.dwi import DwiStudy  # @Reimport

        PipelineTester.generate_test_data(
            DwiStudy, op.join(args.data_dir, 'dwi2'), 'TESTBANANADWI2',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 'dwi2-work'),
            include=['wm_odf'],
            reprocess=False, repo_depth=1, modules_env=True,
            clean_work_dir=(not args.dont_clean_work_dir))

    if 'dwi3' in args.generate:
        from banana import MultiStudy, MultiStudyMetaClass, SubStudySpec
        from banana.study.mri.dwi import DwiStudy  # @Reimport
        from banana.study.mri.t1 import T1Study  # @Reimport

        class DwiT1Study(MultiStudy, metaclass=MultiStudyMetaClass):

            add_substudy_specs = [
                SubStudySpec(
                    't1',
                    T1Study,
                    name_map={
                        'coreg_ref': 'dwi_mag_preproc'}),
                SubStudySpec(
                    'dwi',
                    DwiStudy,
                    name_map={
                        'anat_5tt': 't1_five_tissue_type',
                        'anat_fs_recon_all': 't1_fs_recon_all'})]

        PipelineTester.generate_test_data(
            DwiT1Study, op.join(args.data_dir, 'dwi3'), 'TESTBANANADWI3',
            in_server=None, out_server='https://mbi-xnat.erc.monash.edu.au',
            work_dir=op.join(args.data_dir, 'dwi3-work'),
            include=['dwi_connectome'],
            reprocess=False, repo_depth=0, modules_env=True,
            clean_work_dir=(not args.dont_clean_work_dir))
